import troposphere.kms as kms

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class KMSKey(kms.Key):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.KeyPolicy = {
            'Version': '2012-10-17',
            'Id': 'key-default-1',
            'Statement': [
                {
                    'Action': 'kms:*',
                    'Effect': 'Allow',
                    'Principal': {
                        'AWS': Sub('arn:aws:iam::${AWS::AccountId}:root')
                    },
                    'Resource': '*',
                    'Sid': 'Enable IAM User Permissions'
                },
            ]
        }


class KMSAliasParameterStore(kms.Alias):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.AliasName = Sub('alias/parameter_store_key')
        self.TargetKeyId = Ref('KMSKeyParameterStore')


def KMS_Keys(key):
    # Resources
    R_KeyParameterStore = KMSKey('KMSKeyParameterStore')

    R_AliasParameterStore = KMSAliasParameterStore(
        'KMSAliasParameterStore')

    # Outputs
    O_ParameterStore = Output('KeyParameterStore')
    O_ParameterStore.Value = Sub('${KMSKeyParameterStore.Arn}')
    O_ParameterStore.Export = Export('KeyParameterStore')

    add_obj([
        R_KeyParameterStore,
        R_AliasParameterStore,
        O_ParameterStore])



import troposphere.servicediscovery as srvd

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class ServiceDiscoveryPublicDnsNamespace(srvd.PublicDnsNamespace):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Description = 'Service Discovery'
        self.Name = Sub(f'find.{cfg.HostedZoneNameEnv}')


def SRVD_ServiceDiscoveryRES(key):
    # Resources
    R_PublicDnsNamespace = ServiceDiscoveryPublicDnsNamespace(
        'PublicDnsNamespace')

    # Outputs
    O_PublicDnsNamespace = Output(
        'ServiceDiscoveryPublicDnsNamespaceId')
    O_PublicDnsNamespace.Value = Ref(
        'PublicDnsNamespace')
    O_PublicDnsNamespace.Export = Export(
        'ServiceDiscoveryPublicDnsNamespaceId')

    add_obj([
        R_PublicDnsNamespace,
        O_PublicDnsNamespace])



import troposphere.logs as lgs

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class LogsLogGroup(lgs.LogGroup):
    def __init__(self, title, **kwargs):
        super(LogsLogGroup, self).__init__(title, **kwargs)
        self.LogGroupName = get_endvalue('LogGroupName')
        self.RetentionInDays = get_endvalue('LogRetentionInDays')


def LGS_LogGroup(key):
    R_Group = LogsLogGroup('LogsLogGroup')

    add_obj([
        R_Group])



import troposphere.cloudwatch as clw

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


def CW_Alarms(key):
    for n, v in getattr(cfg, key).items():
        # be carefull that value is enabled in loadbalancing too
        if not v['IBOXENABLED']:
            continue
        resname = f'{key}{n}'
        # resources
        r_Alarm = clw.Alarm(resname)
        auto_get_props(r_Alarm)
        if hasattr(r_Alarm, 'Metrics'):
            r_Alarm.Period = Ref('AWS::NoValue')

        add_obj(r_Alarm)



import troposphere.awslambda as lbd

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj, import_lambda)
from .iam import IAMRoleLambdaBase


class LambdaPermission(lbd.Permission):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Action = 'lambda:InvokeFunction'


class LambdaPermissionEvent(LambdaPermission):
    def __init__(self, title, key, source, **kwargs):
        super().__init__(title, **kwargs)
        self.Principal = 'events.amazonaws.com'
        self.FunctionName = eval(key['Arn'])
        self.SourceArn = GetAtt(source, 'Arn')


class LambdaPermissionS3(LambdaPermission):
    def __init__(self, title, key, source, **kwargs):
        super().__init__(title, **kwargs)
        self.Principal = 's3.amazonaws.com'
        self.FunctionName = key
        self.SourceArn = Sub('arn:aws:s3:::%s' % getattr(cfg, source))


class LambdaPermissionSNS(LambdaPermission):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)
        self.Principal = 'sns.amazonaws.com'
        self.FunctionName = eval(key['Endpoint'])
        self.SourceArn = eval(key['TopicArn'])


class LambdaPermissionApiGateway(LambdaPermission):
    def __init__(self, title, name, source, **kwargs):
        super().__init__(title, **kwargs)
        self.Principal = 'apigateway.amazonaws.com'
        self.FunctionName = name
        self.SourceArn = source


class LambdaPermissionLoadBalancing(LambdaPermission):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Principal = 'elasticloadbalancing.amazonaws.com'
        self.FunctionName = name


class LambdaVersion(lbd.Version):
    def __init__(self, title, name, **kwargs):
        super(LambdaVersion, self).__init__(title, **kwargs)
        self.FunctionName = Ref(name)


class LambdaFunction(lbd.Function):
    def setup(self, key, name):
        import_name = key['ImportName'] if 'ImportName' in key else name
        if 'Code' not in key:
            self.Code = lbd.Code()
            try:
                self.Code.ZipFile = Join('', import_lambda(import_name))
            except Exception:
                self.Code.ZipFile = (
                    'print("Use Code parameter in yaml '
                    f'or create file lib/lambas/{import_name}.code '
                    'with lambda code to execute.")')
        auto_get_props(self)
        self.FunctionName = Sub('${AWS::StackName}-${EnvRole}-%s' % name)
        if 'Handler' not in key:
            self.Handler = 'index.lambda_handler'
        try:
            getattr(self, 'Role')
        except Exception:
            self.Role = GetAtt(f'RoleLambda{name}', 'Arn')

        if all(k in key for k in ['SecurityGroupIds', 'SubnetIds']):
            self.VpcConfig = lbd.VPCConfig('')
            auto_get_props(self.VpcConfig, self.title)

        # Variables - skip if atEdge - always set Env, EnvRole
        try:
            key['AtEdge']
        except Exception as e:
            self.Environment = lbd.Environment(
                Variables={
                    'Env': Ref('EnvShort'),
                    'EnvRole': Ref('EnvRole'),
                }
            )
            if 'Variables' in key:
                self.Environment.Variables.update({
                    varname: get_endvalue(f'{self.title}Variables{varname}')
                    for varname in key['Variables']
                })


class LambdaLayerVersionPermission(lbd.LayerVersionPermission):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Action = 'lambda:GetLayerVersion'
        self.Principal = Ref('AWS::AccountId')


def LambdaLayers(obj, resname, i):
    layername = f'{resname}{i}'
    # parameters
    p_Layer = Parameter(layername)
    p_Layer.Description = layername

    add_obj(p_Layer)

    # condition
    add_obj(
        get_condition(layername, 'not_equals', '', mapinlist=(resname, i))
    )

    # output
    o_Layer = Output(layername)
    o_Layer.Value = Select(i, get_endvalue(resname, condition=layername))
    o_Layer.Value = get_endvalue(resname, condition=layername, mapinlist=i)

    return o_Layer.Value

##


def LBD_Lambdas(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        try:
            v['Code']['S3Key']
        except Exception:
            pass
        else:
            s3keyname = f'{resname}CodeS3Key'
            # parameters
            p_S3Key = Parameter(s3keyname)
            p_S3Key.Description = f'S3Key Name for lambda {n} Code'

            add_obj(p_S3Key)

            # outputs
            o_S3Key = Output(s3keyname)
            o_S3Key.Value = get_endvalue(s3keyname)

            add_obj(o_S3Key)

        # resources
        r_Lambda = LambdaFunction(resname)
        r_Lambda.setup(key=v, name=n)

        if 'Enabled' in v:
            # conditions
            add_obj(get_condition(
                resname, 'not_equals', 'None', f'{resname}Enabled'))

            r_Lambda.Condition = resname

        if 'Layers' in v:
            r_Lambda.Layers = []
            for i, j in enumerate(v['Layers']):
                r_Lambda.Layers.append(
                    LambdaLayers(
                        r_Lambda, '%s%s' % (resname, 'Layers'), i)
                )

        if 'Version' in v:
            versionname = f'{resname}Version'
            versionnameA = f'{versionname}A'
            versionnameB = f'{versionname}B'

            # parameters
            p_Version = Parameter(versionname)
            p_Version.Description = (
                'LambdaVersion - change between A/B '
                'to force deploy new version')
            p_Version.AllowedValues = ['', 'A', 'B']
            p_Version.Default = ''

            add_obj(p_Version)

            # conditons
            c_VersionA = {versionnameA: And(
                Condition(resname) if 'Enabled' in v else Equals('1', '1'),
                get_condition(
                    '', 'equals', 'A', versionname, nomap=True),
            )}

            c_VersionB = {versionnameB: And(
                Condition(resname) if 'Enabled' in v else Equals('1', '1'),
                get_condition(
                    '', 'equals', 'B', versionname, nomap=True),
            )}

            c_Version = {versionname: Or(
                Condition(versionnameA),
                Condition(versionnameB),
            )}

            add_obj([
                c_VersionA,
                c_VersionB,
                c_Version])

            # resources
            r_VersionA = LambdaVersion(versionnameA, name=resname)
            r_VersionA.Condition = versionnameA

            r_VersionB = LambdaVersion(versionnameB, name=resname)
            r_VersionB.Condition = versionnameB

            # outputs
            o_Version = Output(versionname)
            o_Version.Value = If(
                versionnameA,
                Ref(versionnameA),
                Ref(versionnameB)
            )
            o_Version.Condition = versionname

            add_obj([
                r_VersionA,
                r_VersionB,
                o_Version])

        if not v.get('SkipRole'):
            # Automatically setup a lambda Role with base permissions.
            r_Role = IAMRoleLambdaBase(f'Role{resname}', key=v)
            if hasattr(r_Lambda, 'Condition'):
                r_Role.Condition = r_Lambda.Condition
            add_obj(r_Role)

        add_obj(r_Lambda)

        if v.get('Export'):
            O_Lambda = Output(resname)
            O_Lambda.Value = GetAtt(resname, 'Arn')
            O_Lambda.Export = Export(resname)

            add_obj(O_Lambda)


def LBD_LayerVersions(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        try:
            v['Content']['S3Key']
        except Exception:
            pass
        else:
            s3keyname = f'{resname}ContentS3Key'
            # parameters
            p_S3Key = Parameter(s3keyname)
            p_S3Key.Description = f'S3Key Name for lambda {n} Content'

            add_obj(p_S3Key)

            # outputs
            o_S3Key = Output(s3keyname)
            o_S3Key.Value = get_endvalue(s3keyname)

            add_obj(o_S3Key)

        # resources
        r_Layer = lbd.LayerVersion(resname)
        auto_get_props(r_Layer)
        r_LayerPermission = LambdaLayerVersionPermission(
            f'LambdaLayerPermission{n}')
        r_LayerPermission.LayerVersionArn = Ref(resname)

        # output
        o_Layer = Output(resname)
        o_Layer.Value = Ref(resname)

        add_obj([
            r_Layer,
            r_LayerPermission,
            o_Layer])


def LBD_Permissions(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        # resources
        r_Permission = LambdaPermission(resname)
        auto_get_props(r_Permission)

        add_obj([
            r_Permission])


def LBD_EventSourceMappings(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        # resources
        r_EventSourceMapping = lbd.EventSourceMapping(resname)
        auto_get_props(r_EventSourceMapping)

        add_obj([
            r_EventSourceMapping])


def LBD_EventInvokeConfig(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        # resources
        r_EventInvokeConfig = lbd.EventInvokeConfig(resname)
        auto_get_props(r_EventInvokeConfig)

        add_obj([
            r_EventInvokeConfig])



import troposphere.iam as iam

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj, SSMParameter)


class IAMUser(iam.User):
    def __init__(self, title, key, name, **kwargs):
        super().__init__(title, **kwargs)

        self.Condition = self.title
        self.UserName = key['UserName']
        self.LoginProfile = iam.LoginProfile(
            Password=GetAtt(f'SSMParameterPassword{name}', 'Value'),
            PasswordResetRequired=True
        )


class IAMGroup(iam.Group):
    def __init__(self, title, key, name, **kwargs):
        super().__init__(title, **kwargs)

        self.Condition = self.title
        self.GroupName = name


class IAMUserToGroupAddition(iam.UserToGroupAddition):
    def __init__(self, title, key, name, **kwargs):
        super().__init__(title, **kwargs)

        # self.Condition = name
        self.GroupName = name


class IAMPolicy(iam.PolicyType):
    def __init__(self, title, key, name, **kwargs):
        super().__init__(title, **kwargs)

        self.PolicyName = name
        auto_get_props(self)
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }


class IAMManagedPolicy(iam.ManagedPolicy):
    def __init__(self, title, key, name, **kwargs):
        super().__init__(title, **kwargs)

        auto_get_props(self)
        self.PolicyDocument = {
            'Version': '2012-10-17',
        }
        self.Description = key['Description']


class IAMPolicyBucketReplica(iam.PolicyType):
    def __init__(self, title, bucket, bucket_name, mapname, key, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title  # Ex. IAMPolicyReplicaBucketElasticSearch
        self.Condition = f'{bucket}Replica'
        self.PolicyName = self.title
        self.PolicyDocument = {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': [
                        's3:GetReplicationConfiguration',
                        's3:ListBucket'
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        Sub(f'arn:aws:s3:::{bucket_name}')
                    ]
                },
                {
                    'Action': [
                        's3:GetObjectVersion',
                        's3:GetObjectVersionAcl',
                        's3:GetObjectVersionTagging',
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        Sub(f'arn:aws:s3:::{bucket_name}/*')
                    ]
                },
                {
                    'Action': [
                        's3:ReplicateObject',
                        's3:ReplicateDelete',
                        's3:ReplicateTags',
                        's3:ObjectOwnerOverrideToBucketOwner',
                    ],
                    'Effect': 'Allow',
                    'Resource': [
                        If(
                            f'{mapname}{n}DestinationBucket',
                            get_subvalue(
                                '${1M}/*',
                                f'{mapname}{n}DestinationBucket'),
                            Ref('AWS::NoValue')
                        ) for n in key
                    ]
                }
            ]
        }
        self.Roles = [
            Ref(f'Role{self.Condition}')  # Ex. RoleBucketImagesReplica
        ]


class IAMInstanceProfile(iam.InstanceProfile):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Path = '/'
        self.Roles = [
            Ref('RoleInstance')
        ]


class IAMRoleIBox(iam.Role):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        # Trick to add a policy from cfg/yml
        # to a role created from code (Ex. LambdaRoles)
        try:
            cfg.IAMPolicyInRole
        except Exception:
            pass
        else:
            for n, v in cfg.IAMPolicyInRole.items():
                if self.title.endswith(n):
                    self.Policies = [
                        IAMPolicyInRole(n, v)
                    ]
                    break


class IAMRole(iam.Role):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)
        auto_get_props(self)
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Effect': 'Allow',
                'Principal': {
                    key['PrincipalType'] if 'PrincipalType' in key else
                    'Service': [get_endvalue(f'{self.title}Principal')]
                },
            }]
        }
        self.Path = '/'


class IAMRoleUser(iam.Role):
    def __init__(self, title, key, resnameuser, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = f'{resnameuser}RoleAccount'
        self.RoleName = key['UserName']
        self.MaxSessionDuration = 43200
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Condition': {
                    'StringEquals': {'aws:username': key['UserName']}
                },
                'Effect': 'Allow',
                'Principal': {
                    'AWS': Sub('arn:aws:iam::${IdAccount}:root',
                               **{'IdAccount': get_endvalue(
                                   f'{resnameuser}RoleAccount')})},
            }]
        }
        self.Path = '/'


class IAMRoleLambdaBase(IAMRoleIBox):
    def __init__(self, title, key, **kwargs):
        super().__init__(title, **kwargs)

        self.Path = '/'
        self.ManagedPolicyArns = [
            'arn:aws:iam::aws:policy/AmazonEC2ReadOnlyAccess',
            'arn:aws:iam::aws:policy/service-role/'
            'AWSLambdaVPCAccessExecutionRole'
        ]
        if 'RoleManagedPolicyArns' in key:
            self.ManagedPolicyArns.extend(
                get_endvalue('RoleManagedPolicyArns', fixedvalues=key))
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Principal': {'Service': [
                    'lambda.amazonaws.com',
                    'edgelambda.amazonaws.com' if 'AtEdge' in key else
                    Ref('AWS::NoValue')
                ]},
                'Effect': 'Allow'
            }]
        }


class IAMRoleBucketReplica(iam.Role):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)

        self.Condition = self.title.replace('Role', '')
        self.Path = '/'
        self.AssumeRolePolicyDocument = {
            'Statement': [{
                'Action': 'sts:AssumeRole',
                'Effect': 'Allow',
                'Principal': {'Service': ['s3.amazonaws.com']},
            }],
            'Version': '2012-10-17'
        }


# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


def IAMPolicyInRole(name, key):
    Policy = iam.Policy('')
    Policy.PolicyName = name
    Policy.PolicyDocument = {'Version': '2012-10-17'}
    Statement = []
    for n, v in key['Statement'].items():
        Statement.append(IAMPolicyStatement(v))
    Policy.PolicyDocument['Statement'] = Statement

    return Policy


def IAMPolicyStatement(key):
    Statement = {
        'Effect': key['Effect'] if 'Effect' in key else 'Allow',
    }

    for k in ['Action', 'NotAction', 'Resource', 'NotResource', 'Condition']:
        if k in key:
            Statement.update({k: get_endvalue(k, fixedvalues=key)})

    return Statement


def IAMPolicyApiGatewayPrivate():
    policy = {
        'Statement': [{
            'Action': 'execute-api:Invoke',
            'Effect': 'Allow',
            'Principal': '*',
            'Resource': 'execute-api:/*',
        }],
        'Version': '2012-10-17'
    }

    return policy


def IAM_Users(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'  # Ex. IAMUserPincoPalla

        # parameters
        p_Password = Parameter(f'PasswordBase{n}')
        p_Password.Description = (
            'Base Password, must be changed at first login')
        p_Password.Default = ''
        p_Password.NoEcho = True

        add_obj(p_Password)

        # conditions
        c_Enabled = get_condition(
            resname, 'not_equals', 'None', f'{resname}Enabled')

        c_RoleAccount = get_condition(
            f'{resname}RoleAccount', 'not_equals', 'None')

        add_obj([
            c_Enabled,
            c_RoleAccount])

        ManagedPolicyArns = []
        RoleGroups = []
        if 'RoleGroups' in v:
            for m, w in v['RoleGroups'].items():
                condname = f'{resname}RoleGroups{m}'
                # conditions
                add_obj(get_condition(condname, 'not_equals', 'None'))

                # resources
                RoleGroups.append(
                    If(
                        condname,
                        m,
                        Ref('AWS::NoValue')
                    )
                )

                try:
                    policy_arns = cfg.IAMGroup[m]['ManagedPolicyArns']
                except Exception:
                    pass
                else:
                    for p in policy_arns:
                        ManagedPolicyArns.append(
                            If(
                                condname,
                                ImportValue(f'IAMPolicy{p}'),
                                Ref('AWS::NoValue')
                            )
                        )

        # resources
        r_Role = IAMRoleUser(f'IAMRole{n}', key=v, resnameuser=resname)
        r_Role.ManagedPolicyArns = ManagedPolicyArns

        r_User = IAMUser(resname, key=v, name=n)
        r_User.Groups = RoleGroups

        r_SSMParameter = SSMParameter(f'SSMParameterPassword{n}')
        r_SSMParameter.Condition = resname
        r_SSMParameter.Name = Sub(
            '/iam/PasswordBase/${UserName}',
            **{'UserName': get_endvalue(f'{resname}UserName')})
        r_SSMParameter.Value = Ref(f'PasswordBase{n}')
        r_SSMParameter.AllowedPattern = '^[^ ]{16,}$'

        add_obj([
            r_User,
            r_Role,
            r_SSMParameter])


def IAM_UserToGroupAdditions(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'  # Ex. IAMUserToGroupAdditionBase

        Users = []
        for m, w in v['User'].items():
            condname = f'{resname}User{m}'
            # conditions
            add_obj(get_condition(condname, 'not_equals', 'None'))

            Users.append(
                If(
                    condname,
                    # for user defined in the same yaml file
                    # Ref(f'IAMUser{m}') if m in cfg.IAMUser
                    # else get_endvalue(condname),
                    get_endvalue(condname),
                    Ref('AWS::NoValue')
                )
            )

        # resources
        r_GroupAdd = IAMUserToGroupAddition(
            f'IAMUserToGroupAddition{n}', key=v, name=n)
        r_GroupAdd.Users = Users

        add_obj([
            r_GroupAdd])


def IAM_Groups(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'  # Ex. IAMGroupBase

        # conditions
        c_Enabled = get_condition(
            resname, 'not_equals', 'None', f'{resname}Enabled')
        add_obj(c_Enabled)

        ManagedPolicyArns = []
        for m in v['ManagedPolicyArns']:
            if m.startswith('arn'):
                ManagedPolicyArns.append(m)
            elif m.startswith('Ref('):
                ManagedPolicyArns.append(eval(m))
            else:
                ManagedPolicyArns.append(ImportValue(f'IAMPolicy{m}'))

        # resources
        r_Group = IAMGroup(resname, key=v, name=n)
        r_Group.ManagedPolicyArns = ManagedPolicyArns

        add_obj([
            r_Group])


def IAM_Policies(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'  # Ex. IAMPolicyLambdaR53RecordInstanceId
        Statement = []
        for m, w in v['Statement'].items():
            Statement.append(IAMPolicyStatement(w))

        if 'Type' in v and v['Type'] == 'Managed':
            r_Policy = IAMManagedPolicy(resname, key=v, name=n)
        else:
            r_Policy = IAMPolicy(resname, key=v, name=n)
        r_Policy.PolicyDocument['Statement'] = Statement

        add_obj(r_Policy)

        # outputs
        if v.get('Export'):
            o_Policy = Output(resname)
            o_Policy.Value = Ref(resname)
            o_Policy.Export = Export(resname)

            add_obj(o_Policy)


def IAM_Roles(key):
    # Resources
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'  # Ex. RoleECSService
        r_Role = IAMRole(resname, key=v)

        # Add embedded policies if present
        if 'Policies' in v:
            Policies = []
            for p, w in v['Policies'].items():
                Policies.append(IAMPolicyInRole(p, w))
            r_Role.Policies = Policies

        add_obj(r_Role)

        # outputs
        if v.get('Export'):
            o_Role = Output(resname)
            o_Role.Value = GetAtt(resname, 'Arn')
            o_Role.Export = Export(resname)
            if 'Condition' in v:
                o_Role.Condition = v['Condition']

            add_obj(o_Role)



import troposphere.ec2 as ec2

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class SecurityGroup(ec2.SecurityGroup):
    # troposphere ec2 is quite old and SecurityGroupIngress is only a list
    # without the obj type, this break auto_get_props, fix it overriding
    props = {
        'GroupName': (str, False),
        'GroupDescription': (str, True),
        'SecurityGroupEgress': (list, False),
        'SecurityGroupIngress': ([ec2.SecurityGroupRule], False),
        'VpcId': (str, False),
        'Tags': ((Tags, list), False),
    }

    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.VpcId = get_expvalue('VpcId')


class SecurityGroupIngress(ec2.SecurityGroupIngress):
    IpProtocol = 'tcp'


class SecurityGroupRule(ec2.SecurityGroupRule):
    IpProtocol = 'tcp'


class SecurityGroupIngressInstanceELBPorts(SecurityGroupIngress):
    def __init__(self, title, listener, **kwargs):
        super().__init__(title, **kwargs)
        name = self.title  # Ex. SecurityGroupIngressListeners1
        self.FromPort = get_endvalue(f'{listener}InstancePort')
        self.ToPort = get_endvalue(f'{listener}InstancePort')
        self.SourceSecurityGroupId = Ref('SecurityGroupLoadBalancer')
        self.GroupId = GetAtt('SecurityGroupInstancesRules', 'GroupId')


class SecurityGroupEcsService(SecurityGroup):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = 'NetworkModeAwsVpc'
        self.GroupDescription = 'Enable access to Service'
        self.SecurityGroupIngress = []


class SecurityGroupRuleEcsService(SecurityGroupRule):
    def __init__(self, scheme, **kwargs):
        super().__init__(**kwargs)
        self.SourceSecurityGroupId = get_expvalue(
            f'SecurityGroupLoadBalancerApplication{scheme}')
        self.FromPort = get_endvalue('ContainerDefinitions1ContainerPort')
        self.ToPort = self.FromPort


def SG_SecurityGroupsExtra(Out_String, Out_Map):
    # if key SecurityGroups not present,
    # set and empty SecurityGroups list and output (not added)
    try:
        cfg.SecurityGroups
    except Exception:
        return []

    # Parameters
    P_SecurityGroups = Parameter('SecurityGroups')
    P_SecurityGroups.Description = (
        'SecurityGroups List Extra - '
        f'{SECURITY_GROUPS_DEFAULT} for default based on env/role')
    P_SecurityGroups.AllowedPattern = (
        r'^(\w*,\w*){%s}$' % (MAX_SECURITY_GROUPS - 1))
    P_SecurityGroups.Default = SECURITY_GROUPS_DEFAULT

    add_obj([
        P_SecurityGroups])

    SecurityGroups = []

    for n in range(MAX_SECURITY_GROUPS):
        name = f'SecurityGroup{n}'  # Ex SecurityGroup1
        value = Select(
            n, Split(',', get_endvalue('SecurityGroups')))
        outnamename = f'SecurityGroupName{n}'
        outvaluename = f'SecurityGroupValue{n}'

        # conditions
        add_obj({name: Not(get_condition(
            '', 'equals', 'None', Select(n, Split(',', 'SecurityGroups'))))})

        SecurityGroups.append(If(
            name,
            get_expvalue(value, prefix='SecurityGroup'),
            Ref('AWS::NoValue')))

        # outputs
        Out_String.append('${%s}=${%s}' % (outnamename, outvaluename))
        Out_Map.update({
            outnamename: value,
            outvaluename: If(
                name,
                get_expvalue(value, prefix='SecurityGroup'),
                'None')})

    # Outputs
    O_SecurityGroups = Output('SecurityGroups')
    O_SecurityGroups.Value = Sub(','.join(Out_String), **Out_Map)

    add_obj([
        O_SecurityGroups])

    return SecurityGroups


def SG_SecurityGroupsEC2():
    Out_String = ['Rules=${SecurityGroupInstancesRules}']
    Out_Map = {
        'SecurityGroupInstancesRules': {
            'Ref': 'SecurityGroupInstancesRules'}}

    # Resources
    R_SGInstance = SecurityGroup(
        'SecurityGroupInstancesRules')
    R_SGInstance.GroupDescription = (
        'Enable Access from LoadBalancer to Instances '
        'and between Instances')

    add_obj([
        R_SGInstance])

    return SG_SecurityGroupsExtra(Out_String, Out_Map)


def SG_SecurityGroupsECS():
    Out_String = ['Service=${SecurityGroupEcsService}']
    Out_Map = {
        'SecurityGroupEcsService': {
            'Ref': 'SecurityGroupEcsService'}}

    SecurityGroups = SG_SecurityGroupsExtra(Out_String, Out_Map)
    # add Condition to Output created by SG_SecurityGroupsExtra
    try:
        cfg.Outputs['SecurityGroups'].Condition = 'NetworkModeAwsVpc'
    except Exception:
        pass

    return SecurityGroups


def SG_SecurityGroupsTSK():
    Out_String = []
    Out_Map = {}

    SecurityGroups = SG_SecurityGroupsExtra(Out_String, Out_Map)
    # add Condition to Output created by SG_SecurityGroupsExtra
    try:
        cfg.Outputs['SecurityGroups'].Condition = 'NetworkModeAwsVpc'
    except Exception:
        pass

    return SecurityGroups


def SG_SecurityGroupRules(groupname, ingresses):
    SecurityGroup_Rules = []
    kwargs = {}

    # Trick to populate SecurityGroupIngress using cfg.Listeners
    if ingresses:
        # use SecurityGroupIngress
        prefix = f'{groupname}SecurityGroupIngress'
        use_listener = False
        cond_key = 'CidrIp'
    else:
        # use cfg.Listeners
        prefix = 'Listeners'
        ingresses = cfg.Listeners
        use_listener = True
        cond_key = 'Access'

    for n, v in ingresses.items():
        if use_listener:
            # Trick to populate SecurityGroupIngress using cfg.Listeners
            v['CidrIp'] = v['Access']
            v['FromPort'] = v['LoadBalancerPort']
            v['ToPort'] = v['LoadBalancerPort']
            kwargs = {'rootdict': v}

        resname = f'{prefix}{n}'
        allowed_ip = v.get('CidrIp') == 'AllowedIp'
        allowed_ip_or_public = v.get('AllowedIpOrPublic')
        if allowed_ip:
            for m, w in cfg.AllowedIp.items():
                r_SGRule = SecurityGroupRule(resname)
                auto_get_props(r_SGRule, **kwargs)
                auto_get_props(r_SGRule, f'AllowedIp{m}')
                SecurityGroup_Rules.append(If(
                    f'AllowedIp{m}',
                    r_SGRule,
                    Ref('AWS::NoValue')))

        if not allowed_ip or allowed_ip_or_public:
            r_SGRule = SecurityGroupRule(resname)
            auto_get_props(r_SGRule, **kwargs)

            if allowed_ip and allowed_ip_or_public:
                r_SGRule.CidrIp = '0.0.0.0/0'
                # condition
                c_Public = get_condition(
                    f'{resname}Public', 'equals', '0.0.0.0/0',
                    f'{resname}{cond_key}')
                add_obj(c_Public)
                r_SGRule = If(
                    f'{resname}Public', r_SGRule, Ref('AWS::NoValue'))

            SecurityGroup_Rules.append(r_SGRule)

    return SecurityGroup_Rules


def SG_SecurityGroup(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'

        # resources
        r_SG = SecurityGroup(resname)
        auto_get_props(r_SG)
        try:
            ingresses = v['SecurityGroupIngress']
        except Exception:
            pass
        else:
            r_SG.SecurityGroupIngress = SG_SecurityGroupRules(
                resname, ingresses)

        try:
            outname = v['OutputName']
        except Exception:
            outname = resname
        else:
            outname = f'{key}{outname}'

        # outputs
        o_SG = Output(outname)
        o_SG.Value = GetAtt(resname, 'GroupId')
        if v.get('Export'):
            o_SG.Export = Export(outname)

        add_obj(r_SG)
        # add output only if not already present (can be created by IBOXOUTPUT)
        try:
            cfg.Outputs[outname]
        except Exception:
            add_obj(o_SG)


def SG_SecurityGroupIngresses(key):
    for n, v in getattr(cfg, key).items():
        if not v.get('IBOXENABLED', True):
            continue
        resname = f'{key}{n}'
        try:
            allowed_ip = (v['CidrIp'] == 'AllowedIp')
        except Exception:
            pass
        else:
            if allowed_ip:
                for m, w in cfg.AllowedIp.items():
                    r_SGI = SecurityGroupIngress(f'{resname}{m}')
                    auto_get_props(r_SGI, resname)
                    auto_get_props(r_SGI, f'AllowedIp{m}')
                    r_SGI.Condition = f'AllowedIp{m}'
                    add_obj(r_SGI)
                continue

        r_SGI = SecurityGroupIngress(resname)
        auto_get_props(r_SGI)
        add_obj(r_SGI)



import yaml
import yaml.constructor
import sys
import os
import copy
import json
import logging
from collections.abc import Mapping
from pprint import pprint, pformat

from . import cfg


class Loader(yaml.Loader):
    def __init__(self, stream):
        # This way for include relative to file with include statement
        self._root_current = os.path.split(stream.name)[0]
        # This way for include BASE relative on BASE dir
        self._root_base = os.path.join(CFG_FILE_INT, 'BASE')
        # This way for include relative on BASE EXT dir
        self._root_base_ext = os.path.join(CFG_FILE_EXT, 'BASE')
        # This way for include relative on brand EXT dir
        self._root_brand_ext = os.path.join(CFG_FILE_EXT, brand)

        # try to find out if source file is in a subfolder
        if self._root_base in self._root_current:
            suffix = self._root_current.replace(self._root_base, '')
        elif self._root_base_ext in self._root_current:
            suffix = self._root_current.replace(self._root_base_ext, '')
        else:
            suffix = self._root_current.replace(self._root_brand_ext, '')

        # self.root_current_suffix will be empty if source file is in root
        self.root_current_suffix = os.path.basename(suffix)
        if self.root_current_suffix.upper() in cfg.STACK_TYPES:
            # set to empty even if UPPERCASE subfolders is in cfg.STACK_TYPES
            # so that they are not searched for include files
            self.root_current_suffix = ''

        self.stream = stream
        super(Loader, self).__init__(stream)
        Loader.add_constructor('!include', Loader.include)
        Loader.add_constructor('!import', Loader.include)
        Loader.add_constructor('!exclude', Loader.exclude)

    def exclude(self, node):
        global exclude_list
        global include_list
        if isinstance(node, yaml.ScalarNode):
            exclude_list.append(node)
        elif isinstance(node, yaml.SequenceNode):
            for filename in self.construct_sequence(node):
                exclude_list.append(filename)

    def include(self, node):
        if isinstance(node, yaml.ScalarNode) and node not in (
                exclude_list + include_list):
            # a yml file can be included only once
            include_list.append(node)
            return self.extractFile(
                self.construct_scalar(node), self._root_current)

        elif isinstance(node, yaml.SequenceNode):
            result = []
            for filename in self.construct_sequence(node):
                if filename in exclude_list + include_list:
                    continue
                include_list.append(filename)

                # search for include in files in roots
                for path in [self._root_base,
                             self._root_base_ext,
                             self._root_brand_ext]:
                    result.append(self.extractFile(filename, path))
                    if self.root_current_suffix:
                        # search in subfolder too
                        result.append(
                            self.extractFile(filename, os.path.join(
                                path, self.root_current_suffix)))
            return result

        elif isinstance(node, yaml.MappingNode):
            result = {}
            for k, v in self.construct_mapping(node).items():
                if k in exclude_list + include_list:
                    continue
                include_list.append(k)
                result[k] = self.extractFile(v, self._root_current)
            return result

        else:
            print('Error:: unrecognised node type in !include statement')
            raise yaml.constructor.ConstructorError

    def extractFile(self, filename, root):
        filepath = os.path.join(root, filename)
        try:
            with open(filepath, 'r') as f:
                return yaml.load(f, Loader)
        except IOError:
            pass


def replace_not_allowed_char(s):
    key = str(s)
    for s, w in cfg.CLF_PATH_PATTERN_REPLACEMENT.items():
        key = key.replace(s, w)

    return int(key) if key.isdigit() else key


def gen_dict_extract(cfg, envs, check_root=None):
    global enforce_list
    if hasattr(cfg, 'items'):
        # This method allow to delete items from a dictionary
        # while iterating over it
        for k in list(cfg):
            v = cfg[k]
            # for final values
            if (isinstance(v, (str, int, list)) and
                    not k.startswith('IBoxLoader')):
                if k in enforce_list:
                    continue
                elif k.endswith('_IBOXENFORCE'):
                    k = k.replace('_IBOXENFORCE', '')
                    enforce_list.append(k)
                yield {k: v}
            # for empty dict in common.yml
            if isinstance(v, dict) and len(v) == 0:
                yield {k: v}
            # for recursively descending in env/region role dict.
            # List is needed for IBoxLoader include list.
            if k in envs and isinstance(v, (dict, list)):
                if k in ['IBoxLoader', 'IBoxLoaderAfter']:
                    # IBoxLoader included dict processed by
                    # gen_dict_extract have check_root = True
                    kwargs = {'check_root': True}
                else:
                    kwargs = {}
                try:
                    # after descending in env main key
                    # (not the one nested under region) delete key
                    # this way when envs include both (env and region)
                    # i do not process it again
                    if k in list(RP_base.keys()):
                        del cfg[k]
                except Exception:
                    pass
                for result in gen_dict_extract(v, envs, **kwargs):
                    yield result
            if check_root:
                # If iam here means i do not have yet encountered a envs key in
                # an IBoxLoader included dict, so skip properties until i find
                continue
            # for recursively descending in dict not in RP_base_keys
            # (env/region/envrole/stacktype)
            # (final key is the concatenation of traversed dict keys)
            if k not in RP_base_keys and isinstance(v, dict):
                for j, w in v.items():
                    for result in gen_dict_extract({f'{k}{j}': w}, envs):
                        yield result
    if isinstance(cfg, list):
        for n in cfg:
            for result in gen_dict_extract(n, envs, check_root=check_root):
                yield result


def my_merge_dict(basedict, workdict):
    if isinstance(workdict, (str, list)):
        return workdict
    if len(basedict) > 0:
        sumdict = dict(list(basedict.items()) + list(workdict.items()))
    else:
        basedict.update(workdict)
        return dict(list(workdict.items()))

    for k in sumdict.keys():
        try:
            is_dict = isinstance(workdict[k], dict)
            is_map = isinstance(basedict[k], Mapping)
        except Exception:
            is_dict = False
            is_map = False

        # Trick to be able to add an element
        # to a previuosly created/defined list.
        # I can centralize some general iampolicy in ecs-cluster and
        # attach them to RoleInstance, adding element to ManagedPolicyArns,
        # preserving the base ones
        try:
            ibox_add_to_list = (workdict[k][0] == 'IBOXADDTOLIST')
        except Exception:
            ibox_add_to_list = False

        if is_dict and is_map:
            my_merge_dict(basedict[k], workdict[k])
        elif ibox_add_to_list:
            del workdict[k][0]
            basedict[k] += workdict[k]
        elif k in workdict:
            basedict[k] = workdict[k]
        else:
            pass

    return basedict


def get_RP_for_envs(value):
    RP = {}

    try:
        is_dict = isinstance(value[0], dict)
    except Exception:
        is_dict = False

    if hasattr(value, 'items'):
        for d, v in value.items():
            RP[d] = get_RP_for_envs(v)
    elif is_dict:
        for d, v in enumerate(value):
            for i, j in v.items():
                # Trick to overwrite a key - needed if it's value is a dict
                # to avoid that previous dict values are merged.
                if str(i).endswith('**'):
                    key = i.replace('**', '')
                    RP[key] = get_RP_for_envs(j)
                    continue
                # CF Mapping allow for index only alfanumeric char,
                # this way i can specify more "clear" name
                # for index in CloudFormation behaviours
                key = replace_not_allowed_char(i)
                # RP[key] already exist as a dict, try merging
                if key in RP and isinstance(RP[key], dict):
                    RP[key] = my_merge_dict(RP[key], get_RP_for_envs(j))
                else:
                    RP[key] = get_RP_for_envs(j)
    elif isinstance(value, list):
        RP = list(value)
    else:
        RP = value

    return RP


def read_yaml(file_type, brand, base_dir, stacktype=''):
    cfg_file = os.path.join(
        base_dir, brand, stacktype.upper(), f'{file_type}.yml')

    try:
        with open(cfg_file, 'r') as ymlfile:
            cfg = yaml.load(ymlfile, Loader=Loader)

            return cfg
    except IOError:
        return {}


def parse_cfg(cfg, envs=[]):
    parsed_cfg = {}

    for value in gen_dict_extract(cfg, envs):
        for k, v in value.items():
            try:
                if isinstance(v[0], dict):
                    parsed_cfg[k] = parsed_cfg[k] + v
                else:
                    raise
            except Exception:
                parsed_cfg[k] = v

    return parsed_cfg


def merge_cfg(cfgs, cfg_key, list_base=None):
    RP_list = copy.deepcopy(list_base) if list_base else []

    for cfg, v in cfgs.items():
        for c in v:
            keys = ['IBoxLoader', 'IBoxLoaderAfter'] + cfg_key[cfg]
            parsed_cfg = parse_cfg(c, keys)
            RP_list.append(parsed_cfg)

    return RP_list


def prepend_base_cfgs(cfg_cmm):
    base_cfgs = {}
    key_names = {}
    for c in cfg_cmm:
        if isinstance(c, dict):
            for n in list(c):
                # need to delete items while iterating
                v = c[n]
                if n in cfg.BASE_CFGS:
                    if n not in base_cfgs:
                        base_cfgs[n] = {}
                    if n not in key_names:
                        key_names[n] = []
                    base_key_value = cfg.BASE_CFGS[n]
                    for k in list(v):
                        try:
                            base_value = k[base_key_value]
                        except Exception:
                            if isinstance(base_key_value, dict):
                                # use dict value in cfg.BASE_CFGS
                                base_cfgs[n] = base_key_value
                            key_names[n].extend(list(k.keys()))
                        else:
                            # use IBOXBASE (or other value in cfg.BASE_CFGS)
                            for j, w in base_value.items():
                                j_current = base_cfgs[n].get(j)
                                if j_current and isinstance(j_current, list):
                                    # if property is a list extend it
                                    base_cfgs[n][j].extend(w)
                                else:
                                    base_cfgs[n][j] = w
                            v.remove(k)
                    if not c[n]:
                        del c[n]

    for n, v in key_names.items():
        values = []
        # can be optimized with "set(v):" but change order so for now keep it
        for m in v:
            values.append({m: base_cfgs[n]})
        cfg_cmm.insert(0, {n: values})


def get_RP(cfgs):
    RP = copy.deepcopy(RP_base)

    cfg_key_cmm = {
        'common': ['global'],
        'type': [stacktype, 'global'],
        'role': [envrole, 'global'],
    }

    RP_list = []

    cfg_merge_cmm = merge_cfg(cfgs, cfg_key_cmm)

    # Prepend base config from awsibox/cfg.py BASE_CFGS
    prepend_base_cfgs(cfg_merge_cmm)

    RP['cmm']['cmm'] = get_RP_for_envs(cfg_merge_cmm)

    for env, rvalue in RP.items():
        if env == 'cmm':
            continue

        cfg_key_env = {
            'common': [env],
            'type': [env],
            'role': [env],
        }

        cfg_merge_env = merge_cfg(cfgs, cfg_key_env)

        for region in rvalue.keys():

            cfg_key_region = {
                'common': [region],
                'type': [region],
                'role': [region],
            }

            cfg_key_env_region = {
                'common': [env, region],
                'type': [env, region],
                'role': [env, region],
            }

            cfg_merge_region = merge_cfg(
                cfgs, cfg_key_region, cfg_merge_env)
            cfg_merge_env_region = merge_cfg(
                cfgs, cfg_key_env_region, cfg_merge_region)

            RP[env][region] = get_RP_for_envs(cfg_merge_env_region)

    return(RP)


def get_RP_base_keys():
    RP_base_keys = ['global']
    for n, v in RP_base.items():
        RP_base_keys.append(n)
        for m, w in v.items():
            RP_base_keys.append(m)

    return RP_base_keys + [stacktype, envrole]


def set_cfg():
    def RP_to_cfg(key):
        if hasattr(key, 'items'):
            for k, v in key.items():
                setattr(cfg, k, v)
                # recursively traverse dict
                # keys name are the concatenation of traversed dict keys
                if isinstance(v, dict):
                    for j, w in v.items():
                        RP_to_cfg({f'{k}{j}': w})

    # for n, v in cfg.RP_cmm.items():
    #     setattr(cfg, n, v)
    RP_to_cfg(cfg.RP_cmm)

    # set generic attribute based on condition:

    # LoadBalancerClassic
    try:
        cfg.LoadBalancerClassic
    except Exception:
        cfg.LoadBalancerClassic = []

    # LoadBalancerApplication
    try:
        cfg.LoadBalancerApplication
    except Exception:
        cfg.LoadBalancerApplication = []

    # LoadBalancer
    cfg.LoadBalancer = None
    for n in ['LoadBalancerClassic', 'LoadBalancerApplication']:
        try:
            getattr(cfg, n)
        except Exception:
            pass
        else:
            cfg.LoadBalancer = True

    # LoadBalancerClassicExternal LoadBalancerClassicInternal
    cfg.LoadBalancerClassicExternal = None
    cfg.LoadBalancerClassicInternal = None
    try:
        cfg.LoadBalancerClassic
    except Exception:
        pass
    else:
        if 'External' in cfg.LoadBalancerClassic:
            cfg.LoadBalancerClassicExternal = True
        if 'Internal' in cfg.LoadBalancerClassic:
            cfg.LoadBalancerClassicInternal = True

    # LoadBalancerApplicationExternal LoadBalancerApplicationInternal
    cfg.LoadBalancerApplicationExternal = False
    cfg.LoadBalancerApplicationInternal = False
    try:
        cfg.LoadBalancerApplication
    except Exception:
        pass
    else:
        if 'External' in cfg.LoadBalancerApplication:
            cfg.LoadBalancerApplicationExternal = True
        if 'Internal' in cfg.LoadBalancerApplication:
            cfg.LoadBalancerApplicationInternal = True

    # RecordSet
    cfg.RecordSetExternal = None
    cfg.RecordSetInternal = None
    if 'External' in cfg.RecordSet:
        cfg.RecordSetExternal = True
    if 'Internal' in cfg.RecordSet:
        cfg.RecordSetInternal = True


def build_RP():
    global envrole
    global stacktype
    global RP_base
    global RP_base_keys
    global brand
    global CFG_FILE_INT
    global CFG_FILE_EXT
    global enforce_list
    global exclude_list
    global include_list

    enforce_list = []
    exclude_list = []
    include_list = []

    CFG_FILE_INT = '%s/cfg' % os.path.dirname(os.path.realpath(__file__))
    CFG_FILE_INT = os.path.normpath(CFG_FILE_INT)

    CFG_FILE_EXT = '%s/cfg' % os.getcwd()
    CFG_FILE_EXT = os.path.normpath(CFG_FILE_EXT)

    envrole = cfg.envrole
    stacktype = cfg.stacktype
    brand = cfg.brand

    RP_base = {
        'cmm': {
            'cmm': {}
        }
    }

    # dynamically populate RP_base Dict from ENV_BASE and cfg.regions
    for n in cfg.ENV_BASE:
        RP_base[n] = {}
        for m in cfg.regions:
            RP_base[n][m] = {}

    cfg.RP_base = RP_base

    cfg_role = [
        read_yaml(envrole, 'BASE', CFG_FILE_INT, stacktype),
        read_yaml(envrole, 'BASE', CFG_FILE_EXT, stacktype),
        read_yaml(envrole, brand, CFG_FILE_EXT, stacktype),
    ]

    cfgs = {
        'common': [
            read_yaml('common', 'BASE', CFG_FILE_INT),
            read_yaml('common', 'BASE', CFG_FILE_EXT),
            read_yaml('common', brand, CFG_FILE_EXT),
        ],
        'type': [
            read_yaml('TYPE', 'BASE', CFG_FILE_INT, stacktype),
            read_yaml('TYPE', 'BASE', CFG_FILE_EXT, stacktype),
            read_yaml('TYPE', brand, CFG_FILE_EXT, stacktype),
        ],
        'role': cfg_role,
    }

    RP_base_keys = get_RP_base_keys()

    RP = get_RP(cfgs)

    # print(RP['dev']['eu-west-1']['CloudFrontCacheBehaviors']
    #    [2]['QueryStringCacheKeys'])
    if cfg.debug:
        print('##########RP#########START#####')
        pprint(RP)
        print('##########RP#########END#######')
        print('##########ENFORCED######START#####')
        pprint(enforce_list)
        print('##########ENFORCED######END#######')
        print('##########EXCLUDED#######START#####')
        pprint(exclude_list)
        print('##########EXCLUDED#######END#######')
        print('##########INCLUDED#######START#####')
        pprint(include_list)
        print('##########INCLUDED#######END#######')

    cfg.RP = RP
    cfg.RP_cmm = RP['cmm']['cmm']

    set_cfg()



import troposphere.waf as waf
import troposphere.wafregional as wafr

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class WAFIPSet(waf.IPSet):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title  # Ex. WafIPSetAwsNat
        self.Name = name  # Ex. AwsNat


class WAFByteMatchSet(waf.ByteMatchSet):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title
        self.Name = name


class WAFRule(waf.Rule):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title  # Ex. WafRuleBlockIPsNotAllowed
        self.Name = name  # Ex. BlockIPsNotAllowed
        self.MetricName = self.Name


class WAFWebACL(waf.WebACL):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = self.title
        self.Name = name
        self.MetricName = self.Name


class WAFGlobalIPSet(WAFIPSet):
    pass


class WAFGlobalByteMatchSet(WAFByteMatchSet):
    pass


class WAFGlobalRule(WAFRule):
    pass


class WAFGlobalWebAcl(WAFWebACL):
    pass


class WAFRegionalIPSet(wafr.IPSet, WAFIPSet):
    pass


class WAFRegionalByteMatchSet(wafr.ByteMatchSet, WAFByteMatchSet):
    pass


class WAFRegionalRule(wafr.Rule, WAFRule):
    pass


class WAFRegionalWebAcl(wafr.WebACL, WAFWebACL):
    pass

# ############################################
# ### START STACK META CLASSES AND METHODS ###
# ############################################


class WAFIPSetDescriptors(waf.IPSetDescriptors, wafr.IPSetDescriptors):
    def __init__(self, title, name, **kwargs):
        super().__init__(title, **kwargs)
        self.Type = 'IPV4'
        self.Value = Select(int(self.title), get_endvalue(f'{name}Ips'))


class WAFPredicates(waf.Predicates, wafr.Predicates):
    def __init__(self, title, name, ptype, wtype, **kwargs):
        super().__init__(title, **kwargs)
        self.Negated = get_endvalue(f'{name}Negated')
        if ptype == 'ByteMatch':
            self.DataId = Ref(
                f'Waf{wtype}ByteMatchSet{self.title}')
            self.Type = 'ByteMatch'
        if ptype == 'IPMatch':
            self.DataId = Ref(
                f'Waf{wtype}IPSet{self.title}')
            self.Type = 'IPMatch'
        if ptype == 'SizeConstraint':
            self.DataId = Ref(
                f'Waf{wtype}SizeConstraintSet{self.title}')
            self.Type = 'SizeConstraint'
        if ptype == 'SqlInjectionMatch':
            self.DataId = Ref(
                f'Waf{wtype}SqlInjectionMatchSet{self.title}')
            self.Type = 'SqlInjectionMatch'
        if ptype == 'XssMatch':
            self.DataId = Ref(
                f'Waf{wtype}XssMatchSet{self.title}')
            self.Type = 'XssMatch'


class WAFAction(waf.Action, wafr.Action):
    pass


class WAFWebACLRule(waf.Rules, wafr.Rules):
    def __init__(self, title, name, index, wtype, **kwargs):
        super().__init__(title, **kwargs)
        self.Action = WAFAction(
            Type=get_endvalue(f'{name}Action')
        )
        self.Priority = index
        self.RuleId = Ref(f'Waf{wtype}Rule{self.title}')


def WAF_condition(condname, mapname, wtype):
    return {condname: And(
        get_condition('', 'not_equals', 'None', f'{mapname}Enabled'),
        Condition('Global') if wtype == 'Global' else Equals('1', '1'),
        Or(
            get_condition('', 'equals', wtype, f'{mapname}WafType'),
            get_condition('', 'equals', 'Common', f'{mapname}WafType')
        )
    )}

# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################


def WAF_IPSets(key, wtype=''):
    for n, v in getattr(cfg, key).items():
        name = key.replace('Waf', wtype)  # Ex. RegionalIPSet
        resname = f'Waf{name}{n}'  # Ex. WafRegionalIPSetAwsNat
        mapname = f'{key}{n}'  # Ex. WafIPSetAwsNat

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        IPSetDescriptors = []
        for i, j in enumerate(v['Ips']):
            IPSetDescriptor = WAFIPSetDescriptors(str(i), name=mapname)
            IPSetDescriptors.append(IPSetDescriptor)

        IPSet = globals()[f'WAF{name}'](resname, name=n)
        IPSet.IPSetDescriptors = IPSetDescriptors

        add_obj(IPSet)


def WAF_ByteMatchSets(key, wtype=''):
    for n, v in getattr(cfg, key).items():
        name = key.replace('Waf', wtype)  # Ex. RegionalByteMatchSet
        resname = f'Waf{name}{n}'
        mapname = f'{key}{n}'

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        r_ByteMatchSet = globals()[f'WAF{name}'](resname, name=n)
        auto_get_props(r_ByteMatchSet, mapname)

        add_obj(r_ByteMatchSet)


def WAF_Rules(key, wtype=''):
    for n, v in getattr(cfg, key).items():
        name = key.replace('Waf', wtype)  # Ex. RegionalIPSet
        resname = f'Waf{name}{n}'  # Ex. WafRegionalIPSetAwsNat
        mapname = f'{key}{n}'  # Ex. WafIPSetAwsNat

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        Predicates = []
        O_Predicates = []
        for m in v['Predicates']:
            predmapname = f'{mapname}Predicates{m}'
            predresname = m
            ptype = v['Type']
            Predicate = WAFPredicates(
                predresname, name=predmapname, ptype=ptype, wtype=wtype)

            Predicates.append(
                If(
                    Predicate.DataId.data['Ref'],
                    Predicate,
                    Ref('AWS::NoValue')
                )
            )

            O_Predicates.append(m)

        Rule = globals()[f'WAF{name}'](resname, name=n)
        Rule.Predicates = Predicates

        add_obj(Rule)

        # outputs
        O_Rule = Output(f'{mapname}{wtype}')
        O_Rule.Condition = resname
        O_Rule.Value = ','.join(O_Predicates)

        add_obj(O_Rule)


def WAF_WebAcls(key, wtype=''):
    for n, v in getattr(cfg, key).items():
        name = key.replace('Waf', wtype)  # Ex. RegionalIPSet
        resname = f'Waf{name}{n}'  # Ex. WafRegionalIPSetAwsNat
        mapname = f'{key}{n}'  # Ex. WafIPSetAwsNat

        # conditions
        add_obj(WAF_condition(resname, mapname, wtype))

        # resources
        Rules = []
        O_Rules = []
        for w, m in enumerate(v['Rules'], start=1):
            rulemapname = f'{mapname}Rules{m}'
            ruleresname = m  # Ex. BlockIPsNotAllowed
            Rule = WAFWebACLRule(
                ruleresname, name=rulemapname, index=w, wtype=wtype)

            Rules.append(Rule)

            O_Rules.append(m)

        WebACL = globals()[f'WAF{name}'](resname, name=n)
        WebACL.Rules = Rules
        WebACL.DefaultAction = WAFAction(Type=v['DefaultAction'])

        add_obj(WebACL)

        # outputs
        O_WebACL = Output(f'{mapname}{wtype}')
        O_WebACL.Condition = resname
        O_WebACL.Value = Sub('${%s} - %s' % (resname, ','.join(O_Rules)))

        add_obj(O_WebACL)


def WAF_GlobalByteMatchSets(key):
    WAF_ByteMatchSets(key, wtype='Global')


def WAF_GlobalIPSets(key):
    WAF_IPSets(key, wtype='Global')


def WAF_GlobalRules(key):
    WAF_Rules(key, wtype='Global')


def WAF_GlobalWebAcls(key):
    WAF_WebAcls(key, wtype='Global')


def WAF_RegionalByteMatchSets(key):
    WAF_ByteMatchSets(key, wtype='Regional')


def WAF_RegionalIPSets(key):
    WAF_IPSets(key, wtype='Regional')


def WAF_RegionalRules(key):
    WAF_Rules(key, wtype='Regional')


def WAF_RegionalWebAcls(key):
    WAF_WebAcls(key, wtype='Regional')



import troposphere.ecs as ecs

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)
from .securitygroup import (SecurityGroupEcsService,
                            SecurityGroupRuleEcsService)


class ECSLoadBalancer(ecs.LoadBalancer):
    def __init__(self, title, scheme, **kwargs):
        super().__init__(title, **kwargs)
        self.ContainerName = Ref('EnvRole')
        self.ContainerPort = get_endvalue('ContainerDefinitions1ContainerPort')
        self.TargetGroupArn = Ref(f'TargetGroup{scheme}')


class ECSContainerDefinition(ecs.ContainerDefinition):
    def __init__(self, title, key, index, **kwargs):
        super().__init__(title, **kwargs)

        name = self.title  # Ex. ContainerDefinitions1
        auto_get_props(self)

        if len(cfg.ContainerDefinitions) == 1:
            self.Cpu = If(
                'CpuTask',
                get_endvalue('Cpu'),
                get_endvalue(f'{name}Cpu')
            )
            self.Memory = If(
                'LaunchTypeFarGate',
                get_endvalue('Memory'),
                get_endvalue(f'{name}Memory')
            )

        if 'RepoName' in key:
            self.Image = get_subvalue(
                '${1M}.dkr.ecr.${AWS::Region}.amazonaws.com/${2M}:'
                '${EnvApp%sVersion}' % index,
                ['EcrAccount', f'{name}RepoName']
            )
        # use the same EnvApp version for all containers
        elif cfg.RepoName != 'None':
            self.Image = get_subvalue(
                '${1M}.dkr.ecr.${AWS::Region}.amazonaws.com/${2M}:'
                '${EnvApp1Version}',
                ['EcrAccount', 'RepoName']
            )
        elif cfg.Image != 'None':
            self.Image = get_endvalue('Image')

        if 'Name' in key:
            self.Name = get_subvalue('${EnvRole}-${1M}', f'{name}Name')
        else:
            self.Name = Ref('EnvRole')


# ##########################################
# ### END STACK META CLASSES AND METHODS ###
# ##########################################

def ECS_ContainerDefinition():
    Containers = []
    for n, v in cfg.ContainerDefinitions.items():
        Environments = []
        MountPoints = []

        name = f'ContainerDefinitions{n}'  # Ex. ContainerDefinitions1

        # if ContainerDefinitions have RepoName
        # use different EnvApp version
        if n == 1 or 'RepoName' in v:
            nameenvapp = f'EnvApp{n}Version'  # Ex. EnvApp1Version

            # parameters
            EnvApp = Parameter(nameenvapp)
            EnvApp.Description = nameenvapp
            EnvApp.AllowedPattern = '^[a-zA-Z0-9-_.]*$'
            EnvApp.Default = '1'

            add_obj(EnvApp)

            # outputs
            o_EnvAppOut = Output(nameenvapp)
            o_EnvAppOut.Value = Ref(nameenvapp)

            # and use different output for RepoName
            if cfg.RepoName != 'None':
                if 'RepoName' in v:
                    o_Repo = Output(f'{name}RepoName')
                    o_Repo.Value = get_endvalue(f'{name}RepoName')
                else:
                    o_Repo = Output('RepoName')
                    o_Repo.Value = get_endvalue('RepoName')

                add_obj(o_Repo)

            add_obj([
                o_EnvAppOut,
            ])

        # resources
        Container = ECSContainerDefinition(name, key=v, index=n)
        Containers.append(Container)

        # outputs
        EnvValue_Out_String = []
        EnvValue_Out_Map = {}
        for m, w in v['Environment'].items():
            if m.startswith('Env'):
                continue
            envname = f'{name}Environment{m}'
            envkeyname = w['Name']
            # parameters
            p_EnvValue = Parameter(f'{envname}Value')
            p_EnvValue.Description = (
                f'{envkeyname} - empty for default based on env/role')

            # If key NoParam is present skip adding Parameters
            # (usefull as they have a limited max number)
            if 'NoParam' not in w:
                add_obj(p_EnvValue)

            # outputs
            EnvValue_Out_String.append(
                '%s=${%s}' % (envkeyname, envkeyname))
            EnvValue_Out_Map.update({
                envkeyname: get_endvalue(f'{envname}Value')
            })

        o_EnvValueOut = Output(f'{name}Environment')
        o_EnvValueOut.Value = Sub(
            ','.join(EnvValue_Out_String), **EnvValue_Out_Map)

        add_obj(o_EnvValueOut)

    return Containers


def ECS_TaskDefinition(key):
    # Resources
    R_TaskDefinition = ecs.TaskDefinition('TaskDefinitionBase')
    auto_get_props(R_TaskDefinition)
    R_TaskDefinition.ContainerDefinitions = ECS_ContainerDefinition()

    add_obj([
        R_TaskDefinition])


def ECS_Service(key):
    # Resources
    R_SG = SecurityGroupEcsService('SecurityGroupEcsService')
    if cfg.LoadBalancerApplicationExternal:
        SGRule = SecurityGroupRuleEcsService(scheme='External')
        R_SG.SecurityGroupIngress.append(SGRule)

    if cfg.LoadBalancerApplicationInternal:
        SGRule = SecurityGroupRuleEcsService(scheme='Internal')
        R_SG.SecurityGroupIngress.append(SGRule)

    add_obj(R_SG)

    for n, v in getattr(cfg, key).items():
        if not v['IBOXENABLED']:
            continue
        mapname = f'{key}{n}'

        # trick to avoid changing, for now, current service resource name
        if n == 'Spot':
            resname = 'ServiceSpot'
        else:
            resname = 'Service'

        r_Service = ecs.Service(mapname)
        auto_get_props(r_Service)
        r_Service.title = resname

        if cfg.LoadBalancerApplication:
            r_Service.LoadBalancers = []
            r_Service.Role = If(
                'NetworkModeAwsVpc',
                Ref('AWS::NoValue'),
                get_expvalue('RoleECSService'))

        # When creating a service that specifies multiple target groups,
        # the Amazon ECS service-linked role must be created.
        # The role is created by omitting the Role property
        # in AWS CloudFormation
        if (cfg.LoadBalancerApplicationExternal and
                cfg.LoadBalancerApplicationInternal):
            r_Service.Role = Ref('AWS::NoValue')

        if cfg.LoadBalancerApplicationExternal:
            r_Service.LoadBalancers.append(
                ECSLoadBalancer('', scheme='External'))

        if cfg.LoadBalancerApplicationInternal:
            r_Service.LoadBalancers.append(
                ECSLoadBalancer('', scheme='Internal'))

        add_obj(r_Service)


def ECS_Cluster(key):
    # Resources
    R_Cluster = ecs.Cluster('Cluster')

    add_obj([
        R_Cluster,
    ])



import troposphere.codedeploy as cdd

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class CDEc2TagFilters(cdd.Ec2TagFilters):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Key = 'EnvStackName'
        self.Type = 'KEY_AND_VALUE'
        self.Value = Ref('AWS::StackName')


class CDDeploymentGroup(cdd.DeploymentGroup):
    def __init__(self, title, index, **kwargs):
        super().__init__(title, **kwargs)

        appreponame = f'Apps{index}RepoName'
        appenvname = f'EnvApp{index}Version'
        self.Condition = 'DeploymentGroup'
        self.ApplicationName = get_endvalue(appreponame)

        # Uncomment for old behaviour where codedeploy prepare at boot
        # autoscalinggroup's instances (do not work with autospot)
        # self.AutoScalingGroups = If(
        #    'DeployRevision',
        #    [Ref('AutoScalingGroup')],
        #    Ref('AWS::NoValue')
        # )

        # AUTOSPOT - Use Ec2TagFilters
        # to let it work with instances launched by autospot
        self.Ec2TagFilters = If(
            'DeployRevision',
            [CDEc2TagFilters('')],
            Ref('AWS::NoValue')
        )
        self.Deployment = If(
            'DeployRevision',
            cdd.Deployment(
                Revision=cdd.Revision(
                    RevisionType='S3',
                    S3Location=cdd.S3Location(
                        Bucket=Sub(cfg.BucketAppRepository),
                        BundleType='tgz',
                        Key=get_subvalue(
                            '${1M}/${1M}-${%s}.tar.gz'
                            % appenvname, appreponame)
                    )
                )
            ),
            Ref('AWS::NoValue')
        )
        self.DeploymentGroupName = Sub('${AWS::StackName}.${EnvRole}')
        self.ServiceRoleArn = get_expvalue('RoleCodeDeploy', '')


def CD_DeploymentGroup():
    # Conditions
    add_obj([
        {'DeploymentGroup': And(
            Condition('Apps1'),
            get_condition('', 'equals', True, 'DeploymentGroup')
        )},
        {'DeployRevision': Equals(Ref('UpdateMode'), 'CodeDeploy')}])

    # Resources
    # FOR SINGLEAPP CODEDEPLOY
    if len(cfg.Apps) == 1:
        R_DeploymentGroup = CDDeploymentGroup('DeploymentGroup', index=1)

        add_obj(R_DeploymentGroup)


def CD_Applications(key):
    for n, v in getattr(cfg, key).items():
        resname = f'{key}{n}'
        App = cdd.Application(resname)
        App.ApplicationName = get_endvalue(resname)

        add_obj(App)



import troposphere.certificatemanager as crm

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


def CRM_Certificate(key):
    for n, v in getattr(cfg, key).items():
        if not v['IBOXENABLED']:
            continue
        # resources
        resname = f'{key}{n}'
        r_Certificate = crm.Certificate(resname)
        auto_get_props(r_Certificate)
        r_Certificate.Tags = Tags(Name=n)

        # outputs
        o_Certificate = Output(resname)
        o_Certificate.Value = Ref(resname)

        if hasattr(r_Certificate, 'Condition'):
            o_Certificate.Condition = r_Certificate.Condition

        add_obj([
            r_Certificate,
            o_Certificate])


def CRM_CertificateEC2(key):
    if cfg.ListenerLoadBalancerHttpsPort != 'None':
        CRM_Certificate(key)


def CRM_CertificateECS(key):
    if cfg.ListenerLoadBalancerHttpsPort != 443:
        CRM_Certificate(key)


def CRM_CertificateRES(key):
    CRM_Certificate(key)



import troposphere.cloudformation as cfm

from .common import *
from .shared import (Parameter, do_no_override, get_endvalue, get_expvalue,
                     get_subvalue, auto_get_props, get_condition, add_obj,
                     auto_build_obj)


def CFM_Parameters(key):
    auto_build_obj(Parameter(''), getattr(cfg, key))


def CFM_Conditions(key):
    do_no_override(True)
    for n, v in getattr(cfg, key).items():
        c_Condition = {n: eval(v)}

        add_obj(c_Condition)
    do_no_override(False)


def CFM_Mappings(key):
    for n, v in getattr(cfg, key).items():
        c_Mapping = {n: v}

        cfg.Mappings.update(c_Mapping)


def CFM_Outputs(key):
    auto_build_obj(Output(''), getattr(cfg, key))


def CFM_CustomResourceReplicator(key):
    resname = 'CloudFormationCustomResourceStackReplicator'
    # Parameters
    P_ReplicateRegions = Parameter('CCRReplicateRegions')
    P_ReplicateRegions.Description = (
        'Regions where to replicate - None to disable - '
        'empty for default based on env/role')
    P_ReplicateRegions.Type = 'CommaDelimitedList'

    add_obj(P_ReplicateRegions)

    # Resources
    R_Replicator = cfm.CustomResource(resname)

    if 'LambdaCCRStackReplicator' in cfg.Resources:
        R_Replicator.DependsOn = 'IAMPolicyLambdaCCRStackReplicator'
        R_Replicator.ServiceToken = GetAtt('LambdaCCRStackReplicator', 'Arn')
    else:
        R_Replicator.ServiceToken = get_expvalue('LambdaCCRStackReplicator')

    for p, v in cfg.Parameters.items():
        if not p.startswith('Env'):
            value = get_endvalue(p)
        else:
            value = Ref(p)
        setattr(R_Replicator, p, value)

    add_obj(R_Replicator)


def CFM_CustomResourceLightHouse(key):
    resname = 'CloudFormationCustomResourceLightHouse'
    # Parameters
    P_LightHouse = Parameter('CCRLightHouse')
    P_LightHouse.Description = (
        'Enable CustomResource for LightHouse - None to disable - '
        'empty for default based on env/role')

    add_obj(P_LightHouse)

    # Conditions
    C_LightHouse = get_condition(
        resname, 'not_equals', 'None', 'CCRLightHouse')

    add_obj(C_LightHouse)

    # Resources
    R_LightHouse = cfm.CustomResource(resname)
    R_LightHouse.Condition = resname
    R_LightHouse.DependsOn = 'Service'
    R_LightHouse.ServiceToken = get_expvalue('LambdaCCRLightHouse')
    R_LightHouse.EnvRole = Ref('EnvRole')
    R_LightHouse.EnvApp1Version = Ref('EnvApp1Version')
    R_LightHouse.RepoName = get_endvalue('RepoName')

    add_obj(R_LightHouse)


def CFM_CustomResourceFargateSpot(key):
    resname = 'CloudFormationCustomResourceFargateSpot'
    R_FargateSpot = cfm.CustomResource(resname)
    R_FargateSpot.Condition = 'FargateSpot'
    R_FargateSpot.DependsOn = 'ServiceSpot'
    R_FargateSpot.ServiceToken = get_expvalue('LambdaCCRFargateSpot')
    R_FargateSpot.ServiceArn = Ref('ServiceSpot')
    R_FargateSpot.ServiceBase = GetAtt('Service', 'Name')
    R_FargateSpot.ServiceSpot = GetAtt('ServiceSpot', 'Name')
    R_FargateSpot.Cluster = get_expvalue('Cluster', 'ClusterStack')
    R_FargateSpot.ScalingPolicy = Ref('ApplicationAutoScalingScalingPolicyCpu')

    add_obj(R_FargateSpot)



from pprint import pprint, pformat
import argparse

from . import cfg


# parse main argumets
def get_args():
    parser = argparse.ArgumentParser(
        description='Generate AWS CloudFormation Stack Templates')

    # subparser
    subparsers = parser.add_subparsers(help='Desired Action',
                                       required=True, dest='action')

    # parent parser args
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('-R', '--Regions', help='Regions to enable',
                               nargs='+', default=cfg.DEFAULT_REGIONS)
    parent_parser.add_argument('-o', '--Output', help='Output Format',
                               choices=['json', 'cjson', 'yaml'],
                               default='yaml')

    # view parser
    parser_view = subparsers.add_parser(
        'view', parents=[parent_parser],
        help='Generate template and print to standard output')
    parser_view.add_argument('-r', '--EnvRole', help='Stack EnvRole',
                             required=True, type=str)
    parser_view.add_argument('-b', '--Brand', help='Brand',
                             required=True, type=str)
    parser_view.add_argument('-d', '--Debug', help='Show RP Dict',
                             action='store_true')

    # write parser
    parser_write = subparsers.add_parser(
        'write', parents=[parent_parser],
        help=('Generate template '
              'and write to "template/${brand}/$envrole" file'))

    parser_write.add_argument(
        '-b', '--Brands',
        help='Limit generation to EnvRoles defined in the specified Brands',
        nargs='+')
    parser_write.add_argument(
        '-j', '--jobs',
        help='Max concurrent template generation jobs', type=int)

    parser_write.add_argument(
        '-r', '--EnvRoles',
        help='Limit generation to specified EnvRoles',
        nargs='+', default=[])

    parser_write.add_argument(
        '-t', '--StackTypes',
        help='Limit generation to EnvRoles of the specified StackTypes',
        nargs='+', default=[])

    args = parser.parse_args()

    try:
        cfg.debug = args.Debug
    except Exception:
        cfg.debug = None
    cfg.regions = args.Regions

    return args



import troposphere.ec2 as ec2

from .common import *
from .shared import (Parameter, get_endvalue, get_expvalue, get_subvalue,
                     auto_get_props, get_condition, add_obj)


class EC2VPCPeeringConnection(ec2.VPCPeeringConnection):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = 'VPCPeeringConnection'
        self.PeerVpcId = get_expvalue('VpcId-stg', '')
        self.VpcId = get_expvalue('VpcId-dev', '')
        self.Tags = Tags(Name='Dev-Peer-Staging')


class EC2RouteNatGateway(ec2.Route):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.DestinationCidrBlock = '0.0.0.0/0'
        self.NatGatewayId = Ref('NatGateway')
        self.RouteTableId = Ref('RouteTablePrivate')


class EC2RouteInternetGateway(ec2.Route):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.DestinationCidrBlock = '0.0.0.0/0'
        self.GatewayId = Ref('InternetGateway')
        self.RouteTableId = Ref('RouteTablePublic')


class EC2RoutePeeringConnection(ec2.Route):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.Condition = 'VPCPeeringConnection'
        self.VpcPeeringConnectionId = Ref('VPCPeeringConnectionDevStaging')


class EC2RoutePeeringConnectionDev(EC2RoutePeeringConnection):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.DestinationCidrBlock = get_expvalue('VPCCidr-stg', '')
        self.RouteTableId = get_expvalue('RouteTablePrivate-dev', '')


class EC2RoutePeeringConnectionStg(EC2RoutePeeringConnection):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.DestinationCidrBlock = get_expvalue('VPCCidr-dev', '')
        self.RouteTableId = get_expvalue('RouteTablePrivate-stg', '')


class EC2VPCEndpoint(ec2.VPCEndpoint):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.RouteTableIds = [get_expvalue('RouteTablePrivate')]
        self.VpcId = get_expvalue('VpcId')


class EC2VPCEndpointS3(EC2VPCEndpoint):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.ServiceName = Sub('com.amazonaws.${AWS::Region}.s3')


class EC2VPC(ec2.VPC):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.CidrBlock = Ref('VPCCidrBlock')
        self.EnableDnsSupport = True
        self.EnableDnsHostnames = True


class EC2Subnet(ec2.Subnet):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, **kwargs)
        self.AvailabilityZone = Sub('${AWS::Region}%s' % zone.lower())
        self.VpcId = Ref('VPC')


class EC2SubnetPrivate(EC2Subnet):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, zone, **kwargs)
        self.CidrBlock = Ref(f'SubnetCidrBlockPrivate{zone}')
        self.MapPublicIpOnLaunch = False
        self.Tags = Tags(Name=Sub('${VPCName}-Private%s' % zone))


class EC2SubnetPublic(EC2Subnet):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, zone, **kwargs)
        self.CidrBlock = Ref(f'SubnetCidrBlockPublic{zone}')
        self.MapPublicIpOnLaunch = True
        self.Tags = Tags(Name=Sub('${VPCName}-Public%s' % zone))


class EC2RouteTable(ec2.RouteTable):
    def __init__(self, title, **kwargs):
        super().__init__(title, **kwargs)
        self.VpcId = Ref('VPC')


class EC2SubnetRouteTableAssociationPrivate(ec2.SubnetRouteTableAssociation):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, **kwargs)
        self.RouteTableId = Ref('RouteTablePrivate')
        self.SubnetId = Ref(f'SubnetPrivate{zone}')


class EC2SubnetRouteTableAssociationPublic(ec2.SubnetRouteTableAssociation):
    def __init__(self, title, zone, **kwargs):
        super().__init__(title, **kwargs)
        self.RouteTableId = Ref('RouteTablePublic')
        self.SubnetId = Ref(f'SubnetPublic{zone}')
##


def VPC_Endpoint(key):
    # Conditions
    c_VPCEndpoint = get_condition(
        'EC2VPCEndpointS3', 'not_equals', 'None', 'VPCEndpoint')

    add_obj(c_VPCEndpoint)

    # Resources
    R_S3 = EC2VPCEndpointS3('EC2VPCEndpointS3')
    R_S3.Condition = 'EC2VPCEndpointS3'

    add_obj([
        R_S3])


def VPC_VPC(key):
    vpc_net = '10.80'
    o_subnetprivate = []
    o_subnetpublic = []

    # Resources
    R_VPC = ec2.VPC('VPC')
    auto_get_props(R_VPC, f'{key}Base')

    R_RouteTablePrivate = EC2RouteTable('RouteTablePrivate')
    R_RouteTablePrivate.Tags = Tags(Name=Sub('${VPCName}-Private'))

    R_RouteTablePublic = EC2RouteTable('RouteTablePublic')
    R_RouteTablePublic.Tags = Tags(Name=Sub('${VPCName}-Public'))

    R_InternetGateway = ec2.InternetGateway('InternetGateway')
    R_InternetGateway.Tags = Tags(Name=Ref('VPCName'))

    R_VPCGatewayAttachment = ec2.VPCGatewayAttachment(
        'VPCGatewayAttachment')
    R_VPCGatewayAttachment.InternetGatewayId = Ref('InternetGateway')
    R_VPCGatewayAttachment.VpcId = Ref('VPC')

    if cfg.NatGateway != 'None':
        # resources
        R_EIPNat = ec2.EIP('EIPNat')
        R_EIPNat.Domain = 'vpc'

        R_NatGateway = ec2.NatGateway('NatGateway')
        R_NatGateway.AllocationId = GetAtt('EIPNat', 'AllocationId')
        R_NatGateway.SubnetId = Ref('SubnetPublicA')

        R_RouteNatGateway = EC2RouteNatGateway('RouteNatGateway')

        # outputs
        O_EIPNat = Output('EIPNat')
        O_EIPNat.Value = Ref('EIPNat')

        add_obj([
            R_NatGateway,
            R_RouteNatGateway,
            R_EIPNat,
            O_EIPNat])

    R_RouteInternetGateway = EC2RouteInternetGateway(
        'RouteInternetGateway')

    add_obj([
        R_VPC,
        R_RouteTablePrivate,
        R_RouteTablePublic,
        R_InternetGateway,
        R_VPCGatewayAttachment,
        R_RouteInternetGateway])

    for i in range(cfg.AZones['MAX']):
        zone_name = cfg.AZoneNames[i]
        zone_cond = f'Zone{zone_name}'

        # parameters
        p_SubnetCidrBlockPrivate = Parameter(
            f'SubnetCidrBlockPrivate{zone_name}')
        p_SubnetCidrBlockPrivate.Description = (
            f'Ip Class Range for Private Subnet in Zone {zone_name}')
        p_SubnetCidrBlockPrivate.Default = f'{vpc_net}.{i * 16}.0/20'

        p_SubnetCidrBlockPublic = Parameter(
            f'SubnetCidrBlockPublic{zone_name}')
        p_SubnetCidrBlockPublic.Description = (
            f'Ip Class Range for Public Subnet in zone {zone_name}')
        p_SubnetCidrBlockPublic.Default = f'{vpc_net}.{i + 200}.0/24'

        add_obj([
            p_SubnetCidrBlockPrivate,
            p_SubnetCidrBlockPublic])

        # conditions
        c_Zone = {zone_cond: Equals(
            FindInMap(
                'AvabilityZones',
                Ref('AWS::Region'),
                f'Zone{i}'
            ),
            'True')}

        add_obj(c_Zone)

        # resources

        r_SubnetPrivate = EC2SubnetPrivate(
            f'SubnetPrivate{zone_name}', zone=zone_name)
        r_SubnetPrivate.Condition = zone_cond

        r_SubnetPublic = EC2SubnetPublic(
            f'SubnetPublic{zone_name}', zone=zone_name)
        r_SubnetPublic.Condition = zone_cond

        r_SubnetRouteTableAssociationPrivate = (
            EC2SubnetRouteTableAssociationPrivate(
                f'SubnetRouteTableAssociationPrivate{zone_name}',
                zone=zone_name))
        r_SubnetRouteTableAssociationPrivate.Condition = zone_cond

        r_SubnetRouteTableAssociationPublic = (
            EC2SubnetRouteTableAssociationPublic(
                f'SubnetRouteTableAssociationPublic{zone_name}',
                zone=zone_name))
        r_SubnetRouteTableAssociationPublic.Condition = zone_cond

        add_obj([
            r_SubnetPrivate,
            r_SubnetPublic,
            r_SubnetRouteTableAssociationPrivate,
            r_SubnetRouteTableAssociationPublic])

        # outputs
        o_subnetprivate.append(If(
            zone_cond,
            Ref(f'SubnetPrivate{zone_name}'),
            Ref('AWS::NoValue')
        ))

        o_subnetpublic.append(If(
            zone_cond,
            Ref(f'SubnetPublic{zone_name}'),
            Ref('AWS::NoValue')
        ))

    # Outputs
    O_SubnetsPrivate = Output('SubnetsPrivate')
    O_SubnetsPrivate.Value = Join(',', o_subnetprivate)
    O_SubnetsPrivate.Export = Export('SubnetsPrivate')

    O_SubnetsPublic = Output('SubnetsPublic')
    O_SubnetsPublic.Value = Join(',', o_subnetpublic)
    O_SubnetsPublic.Export = Export('SubnetsPublic')

    add_obj([
        O_SubnetsPrivate,
        O_SubnetsPublic])
